/*++ nqs_aboque.c - Network Queueing System
 *
 * $Source$
 *
 * DESCRIPTION:
 *
 *	Abort all running requests in an NQS queue.
 *
 *
 *	Author:
 *	-------
 *	Brent A. Kingsbury, Sterling Software Incorporated.
 *	August 12, 1985.
 *
 *
 * STANDARDS VIOLATIONS:
 *   None.
 *
 * REVISION HISTORY: ($Revision$ $Date$ $State$)
 * $Log$
 *
 *	August 7, 1991	Global batch run and pipe route limits
 *	August 12, 1991	Only SIGKILL signal sent during shutdown sequence
 *	November 22, 1991 Change default signal to send to have the same
 *			behaviour with various shells
 *
 */

#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include "nqs.h"			/* NQS constants and data types */
#include "nqsxvars.h"			/* Global vars */

extern char *asciierrno();		/* Return ASCII errno */
extern int errno;			/* System error number */
extern void nqs_vtimer();		/* Set virtual timer */
extern time_t time();			/* Time in seconds */

static char *info = "I$%s.\n";
static char *problem = "E$Problem sending signal in nqs_aboque.c.\n";

/*** nqs_aboque
 *
 *
 *	void nqs_aboque():
 *	Abort all running requests in an NQS queue.
 */
void nqs_aboque (queue, grace_period, requeue_flag)
struct qhead *queue;			/* Pointer to queue structure */
int grace_period;			/* Number of seconds to wait */
					/* after sending SIGINT to send */
					/* SIGKILL. */
short requeue_flag;			/* Boolean non-zero if aborted */
					/* requests are to be requeued */
{
	void nqs_kill();		/* Send a SIGKILL to all remaining */
					/* processes for each remaining */
					/* running request in queues which */
					/* are being aborted. */

	short sig;			/* Signal to send */
	time_t timeval;			/* Timer value */
	register short reqindex;	/* Index into the Runvars[] array */
					/* of the running structure for */
					/* the request being aborted */
	register struct request *req;	/* Ptr to request structure for */
					/* the request being aborted */

	if (queue->sigkill) {
	    /*
	     *  Multiple pending ABort Queue commands are NOT allowed
	     *  for the same queue.
	     */
	    return;
	}
	sig = SIGINT;
	if (grace_period == 0) sig = SIGKILL;
	else {
	    time (&timeval);
	    timeval += grace_period;
	    nqs_vtimer (&timeval, nqs_kill);	/* Set timer for */
	    queue->sigkill = timeval;		/* SIGKILL. */
	}
	req = queue->runset;
	while (req != (struct request *) 0) {
	    /*
	     *  It looks like this request is still running, though
	     *  it could have finished running without our hearing
	     *  about it just yet.  Send sig to all of its processes.
	     */
	    reqindex = req->reqindex;
	    if ((Runvars + reqindex)->process_family == 0) {
		/*
		 *  The process group/family of the server
		 *  has not yet been reported in....  Queue
		 *  the signal so that it can be sent later
		 *  upon receipt of the process group/family
		 *  packet for this request.
		 */
		if ((req->status & RQF_SIGQUEUED) == 0 ||
		     (Runvars + reqindex)->queued_signal != SIGKILL) {
		    /*
		     *  Queue the signal, unless a SIGKILL
		     *  has already been queued.
		     */
		    (Runvars + reqindex)->queued_signal = sig;
		}
		req->status |= RQF_SIGQUEUED;	/* Queued signal */
		if (requeue_flag) req->status |= RQF_SIGREQUEUE;
		else req->status &= ~RQF_SIGREQUEUE;
	    }
	    else {
		/*
		 *  The process group/family of the server
		 *  is known.  Send the signal.
		 */
#if	SGI | SYS52 | SYS53 | UTS | UNICOS
		if ((grace_period == 0 || requeue_flag == 0) && kill (
			-(Runvars + reqindex)->process_family, sig) == -1) {
#else
#if	BSD42 | BSD43 | ULTRIX
		if ((grace_period == 0 || requeue_flag == 0) && killpg (
			(Runvars + reqindex)->process_family, sig) == -1) {
#else
BAD SYSTEM TYPE
#endif
#endif
		    if (errno != ESRCH) {
			/*
			 *  Something went really wrong!
			 */
			printf (problem);
			printf (info, asciierrno ());
			fflush (stdout);
		    }
		    /*  Else the request is already done! */
		}
		else if (requeue_flag) req->status |= RQF_SIGREQUEUE;
		else req->status &= ~RQF_SIGREQUEUE;
	    }
	    req = req->next;		/* Get the next request in the */
	}				/* running set */
}


/*** nqs_kill
 *
 *
 *	void nqs_kill():
 *
 *	Terminate ALL remaining request processes for the NQS queues that
 *	have been aborted.
 */
static void nqs_kill()
{
	void nqs_killset();

	nqs_killset (Nonnet_queueset);	/* Non-network queues */
	nqs_killset (Net_queueset);	/* Network queues only */
}


/*** nqs_killset
 *
 *
 *	void nqs_killset():
 *
 *	Zealously murder ALL remaining request processes for a set of queues
 *	that have been aborted, and whose SIGKILL times have arrived.  We
 *	sally forth, bandying our bloody axe with the words "SIGKILL" written
 *	upon it.
 */
static void nqs_killset (queue)
register struct qhead *queue;		/* Ptr to a queue structure */
{
	time_t timenow;			/* Time now */
	register short reqindex;	/* Index in the Runvars array of */
					/* running structure allocated for */
					/* the request being aborted */
	register struct request *req;	/* Ptr to request structure for */
					/* the request being aborted */

	time (&timenow);
	/*
	 *  Walk the queue set looking for aborted queues, whose SIGKILL
	 *  time has arrived.
	 */
	while (queue != (struct qhead *) 0) {
	    if (queue->sigkill && queue->sigkill <= timenow) {
		/*
		 *  Swing the axe and cutlass!
		 */
		queue->sigkill = 0;	/* Clear SIGKILL abort flag */
		req = queue->runset;
		while (req != (struct request *) 0) {
		    /*
		     *  It looks like this request is still
		     *  running, though it could have finished
		     *  without our hearing about it just yet.
		     *  Send SIGKILL to all of its processes.
		     */
		    reqindex = req->reqindex;
		    if ((Runvars + reqindex)->process_family == 0) {
			/*
			 *  The process group/family of the
			 *  server has not yet been reported
			 *  in....  Queue the SIGKILL so that
			 *  it can be sent later upon receipt
			 *  of the process group/family
			 *  packet for this request.
			 */
			req->status |= RQF_SIGQUEUED;
			(Runvars + reqindex)->queued_signal = SIGKILL;
		    }
		    else {
			/*
			 *  The process group/family of the
			 *  server is known.  Send SIGKILL.
			 */
#if	SGI | SYS52 | SYS53 | UTS | UNICOS
			if (kill (-(Runvars + reqindex)->process_family,
				  SIGKILL) == -1) {
#else
#if	BSD42 | BSD43 | ULTRIX
			if (killpg ((Runvars + reqindex)->process_family,
				    SIGKILL) == -1) {
#else
BAD SYSTEM TYPE
#endif
#endif
			    if (errno != ESRCH) {
				/*
				 *  Something went really wrong!
				 */
				printf (problem);
				printf (info, asciierrno ());
				fflush (stdout);
			    }
			    /* Else the request is already done! */
			}
		    }
		    req = req->next;	/* Get the next request in the */
		}			/* running set */
	    }
	    else if (queue->sigkill) {
		/*
		 *  This queue has an abort SIGKILL time in the
		 *  future.  Set the virtual timer appropriately.
		 */
		nqs_vtimer (&queue->sigkill, nqs_kill);
	    }
	    queue = queue->next;	/* Get the next queue */
	}
}
