/*++ nqs_bsc.c - Network Queueing System
 *
 * $Source$
 *
 * DESCRIPTION:
 *
 *	This module contains the 4 functions:
 *
 *		bsc_reqcom()
 *		bsc_sched()
 *		bsc_spawn()
 *		bsc_verqueusr()
 *
 *	which control the scheduling, and spawning of NQS batch requests.
 *	This module can be modified to implement appropriate scheduling
 *	algorithms for a particular installation as necessary.
 *
 *
 *	Author:
 *	-------
 *	Brent A. Kingsbury, Sterling Software Incorporated.
 *	August 12, 1985.
 *
 *
 * STANDARDS VIOLATIONS:
 *   None.
 *
 * REVISION HISTORY: ($Revision$ $Date$ $State$)
 * $Log$
 *
 *	July 12, 1991	User-limit for batch queues
 *	August 19, 1991	Verify batch queue user-limit
 *	September 25, 1991 Do not loose the next request when spawning one
 *
 */

#include <stdio.h>
#include "nqs.h"			/* NQS constants and data types */
#include "transactcc.h"			/* Transaction completon codes */
#include "nqsxvars.h"			/* NQS global variables */

extern void udb_qorder();		/* Update queue-ordering */
extern void nqs_spawn();		/* Spawn an NQS request */
extern struct qhead *nqs_fndnnq();
					/* Find non-network queue by name */


/*** bsc_reqcom
 *
 *
 *	void bsc_reqcom():
 *
 *	This function is invoked whenever a running batch request completes
 *	execution.  This function is free to spawn more batch requests from
 *	ANY batch queue.
 *
 *	All queues modified by the invocation of this procedure have
 *	had their corresponding NQS database queue state image updated
 *	(to match the internal queue states) upon return from this
 *	procedure.
 */
void bsc_reqcom (request)
struct request *request;		/* The completed batch request */
{
	void bsc_spawn();

	/*
	 *  For the moment, we do not use any of the knowledge imparted
	 *  to us by the request that just finished execution.  Instead,
	 *  we spawn as many batch requests as we can, within the limits
	 *  configured.
	 */
	bsc_spawn();			/* Spawn all batch requests */
	if (request->queue->v1.batch.qcomplex[0] != (struct qcomplex *) 0)
		dsc_spawn();
}					/* within configured boundaries */


/*** bsc_sched
 *
 *
 *	int bsc_sched():
 *
 *	This function is invoked whenever a batch req must be evaluated
 *	and assigned a priority by the NQS batch req scheduling policies
 *	(which are implemented by this function).
 *
 *	The priority assigned to the req must be in the interval [0..32767].
 *	All batch reqs with priority >= N within a given batch queue, will,
 *	depending upon the precise scheduling criteria defined in:
 *
 *		bsc_spawn()  and
 *		bsc_reqcom()
 *
 *	be spawned before any batch reqs in the same queue with a priority
 *	value < N.
 *	
 *	Returns:
 *		The assigned priority value for the specified batch request.
 */
int bsc_sched (rawreq)
struct rawreq *rawreq;			/* The raw request structure */
					/* describing the new request */
{
	if (rawreq->rpriority == -1) {
		/*
		 *  The user did not specify a priority; assign a
		 *  default value.
		 */
		return (Defbatpri);
	}
	return (rawreq->rpriority);	/* For now, just the intra-queue */
}					/* req priority */


/*** bsc_spawn
 *
 *
 *	void bsc_spawn():
 *
 *	This function is invoked whenever request activity indicates that
 *	it MAY be possible to spawn a batch request.  It is up to the
 *	discretion of the batch request scheduling/spawning algorithm
 *	implemented here to determine whether or not batch req(s) should
 *	be spawned.
 *
 *	All queues modified by the invocation of this procedure have
 *	had their corresponding NQS database queue state image updated
 *	(to match the internal queue states) upon return from this
 *	procedure.
 */
void bsc_spawn()
{
	register struct qhead *queue;	/* Batch queue set walking */
	register int prevbatcount;	/* Prev loop value of Gblbatcount */
	struct request *walkreq;	/* Queue request set walking */
	struct request *nextreq;	/* Next queue request to walk */
	register struct qcomplex *qcomplex;
					/* Queue complex walking */
	register int full;		/* Some qcomplex is full flag */
	register int i;

	/*
	 *  Note that we are very careful to make sure that all batch
	 *  queues with higher priorities that also have batch requests
	 *  that can run, get to spawn first.  This becomes critical when
	 *  the number of batch requests that can run exceeds the maximum
	 *  number of batch requests that are allowed to simultaneously
	 *  execute.
	 */
	if (Shutdown) return;		/* Do not spawn any requests if */
					/* NQS is shutting down */
	queue = Pribatqueset;		/* Prioritized batch queue set */
	prevbatcount = Gblbatcount - 1;	/* Make loop go at least once */
	while (queue != (struct qhead *) 0 &&
	       Maxgblbatlimit > Gblbatcount &&
	       prevbatcount != Gblbatcount) {
		/*
		 *  Spawn as many batch requests as we can for this queue.
		 */
		nextreq = queue->queuedset;
		while (queue->q.v1.batch.runlimit > queue->q.runcount &&
		      (queue->q.status & QUE_RUNNING) &&
		       (walkreq = nextreq) != (struct request *) 0 &&
		       Maxgblbatlimit > Gblbatcount &&
		       prevbatcount != Gblbatcount) {
			full = 0;
			for (i = MAX_COMPLXSPERQ; --i >= 0;) {
				qcomplex = queue->v1.batch.qcomplex[i];
				if (qcomplex == (struct qcomplex *)0) continue;
				if (qcomplex->runlimit <= qcomplex->runcount) {
					full = 1;
					break;
				}
			}
			if (full) break;
			/*
			 *  Find the next request first. After it could be
			 *  too late...
			 */
			nextreq = walkreq->next;
			/*
			 *  Check queue user run limit.
			 */
			if (queue->q.v1.batch.userlimit >
				userrun (walkreq->v1.req.uid, queue)) {
				/*
			 	 *  There is a batch request that can be
				 *  spawned. Try to spawn it.
			 	 *
			 	 *  Note that when spawning a batch request,
			 	 *  subrequests must be created and queued in
			 	 *  the appropriate network queues--which
			 	 *  themselves may have to be created.  It is
			 	 *  therefore possible that the successful
			 	 *  spawning of a batch request would require
			 	 *  more memory, than there is available.  In
			 	 *  such a situation, the spawn quietly fails.
			 	 *  By watching Gblbatcount, this procedure
			 	 *  can tell when such an event has occurred....
			 	 */
				prevbatcount = Gblbatcount;
				nqs_spawn (walkreq, (struct device *) 0);
			}
		}
		if (queue->q.status & QUE_UPDATE) {
			/*
			 *  The database image of this queue must be
			 *  updated.
			 */
			udb_qorder (queue);	/* Update image and clear */
		}				/* the QUE_UPDATE bit */
		queue = queue->v1.batch.nextpriority;	/* Next queue */
	}
}


/*** bsc_verqueusr
 *
 *
 *	long bsc_verqueusr():
 *	Verify user-limit for a queue.
 *
 *	This function is invoked whenever queue activity indicates that
 *	it MAY be possible to spawn a batch request. It is the main part
 *	of the load balancing mechanism. Note that the queue may be
 *	specified by a pointer to its structure or by name (requests
 *	coming through the network).
 *
 *	Returns:
 *		TCML_SUBMITTED if successful. Otherwise TCML_QUEBUSY.
 */
long bsc_verqueusr (uid, queue, que_name)
uid_t uid;					/* User id */
struct qhead *queue;				/* Queue in question */
char *que_name;					/* Queue name in question */
{
	struct qhead *walkque;			/* Batch queue set walking */
	register struct qcomplex *qcomplex;	/* Queue complex walking */
	register int i;

	/*
	 *  Find the queue structure if needed.
	 */
	if ((walkque = queue) == (struct qhead *) 0)
		walkque = nqs_fndnnq (que_name);
	/*
	 *  Is there a free initiator ?
	 */
	if (Maxgblbatlimit > Gblbatcount &&
	    walkque->q.v1.batch.runlimit > (walkque->q.runcount +
		walkque->q.arrivecount + walkque->q.queuedcount) &&
	/*
	 *  All the arriving and queued requests have to be counted as
	 *  running because with the load-balancing mechanism, on the local
	 *  batch queues, all of them should be running...
	 */
	    walkque->q.v1.batch.userlimit > userrun (uid, walkque)) {
		for (i = MAX_COMPLXSPERQ; --i >= 0;) {
			qcomplex = walkque->v1.batch.qcomplex[i];
			if (qcomplex == (struct qcomplex *)0) continue;
			if (qcomplex->runlimit <= qcomplex->runcount)
				return (TCML_QUEBUSY);
		}
		return (TCML_SUBMITTED);
	} else return (TCML_QUEBUSY);
}


/*** userrun
 *
 *      int userrun():
 *      Return the global or queue run count of the specified user.
 *
 */
static int userrun (uid, queue)
uid_t uid;
struct qhead *queue;
{
	register struct request *walkreq;	/* Queue request set walking */
	register int run_count;

	run_count = 0;
	for (walkreq = queue->runset; walkreq != (struct request *) 0;
	     walkreq = walkreq->next) {
		if (walkreq->v1.req.uid == uid)
			run_count++;		/* A match ! */
	}
	return (run_count);
}
