/*++ nqs_complt.c - Network Queueing System
 *
 * $Source$
 *
 * DESCRIPTION:
 *
 *	Complete a local intra-machine NQS transaction, by sending
 *	a transaction completion code to the requesting process.
 *
 *	Let's talk for a moment, shall we?  This module could have
 *	been implemented using named-pipes, or on some systems, by a
 *	socket or socket like connection, or even by datagrams.
 *
 *	However, I'm trying very hard to keep this thing PORTABLE.
 *	Also, this implementation has one principal advantage:
 *
 *		The burden of creating and opening
 *		the file is placed on the client.
 *		If the simple inter-process communication
 *		file cannot be created and opened
 *		by the client, then the client 
 *		simply does not try to perform any
 *		transactions with the NQS daemon.
 *		We find out up front, whether or
 *		not we can talk (as a local client
 *		process), to the local NQS daemon.
 *		If the supporting system is short of
 *		open file table entries, we find this
 *		right away.
 *
 *
 *	This scheme is very sleazy, and I apologize here.  If you
 *	have a better idea that is still portable, then please
 *	step forward and tell me about it!
 *
 *
 *	Author:
 *	-------
 *	Brent A. Kingsbury, Sterling Software Incorporated.
 *	October 8, 1985.
 *
 *
 * STANDARDS VIOLATIONS:
 *   None.
 *
 * REVISION HISTORY: ($Revision$ $Date$ $State$)
 * $Log$
 *
 */

#include "nqs.h"		/* Packet size definition */
#include "nqsxvars.h"		/* NQS external vars */
#include <stdio.h>
#include <errno.h>
#include <signal.h>		/* Defines signals and int (*signal())(); */
#if	BSD42 | BSD43 | ULTRIX
#include <sys/time.h>
#else
#if	UNICOS | SGI | SYS52 | SYS53 | UTS
struct utimbuf {
	time_t acctime;		/* Access time */
	time_t modtime;		/* Modification time */
};
#else
BAD SYSTEM TYPE
#endif
#endif

/*
 *	LMTOFFSET is 0 on all machines except on Apollo where utimes
 *	does not set modifications times =< 1 Jan 1980
 */
#ifdef	DOMAIN_OS
#define LMTOFFSET	315532801L
#else
#define LMTOFFSET	0L
#endif

extern int errno;		/* System call error# */
extern void nqs_abort();	/* NQS abort */
extern time_t time();		/* Get time */

/*** nqs_complt
 *
 *
 *	void nqs_complt():
 *
 *	Complete a local intra-machine NQS transaction, by sending
 *	a transaction completion code to the requesting process.
 */
void nqs_complt (completion_code, pid)
long completion_code;			/* Transaction completion code */
int pid;				/* Process-id of requester */
{
	static char incomplete[]
	= "W$Nqs_complt() unable to send completion code.\n";

#if	BSD42 | BSD43 | ULTRIX
	struct timeval utimbuf [2];	/* Utimes() buffer */
#else
#if	UNICOS | SGI | SYS52 | SYS53 | UTS
	struct utimbuf utimbuf;		/* utime() buffer */
#else
BAD SYSTEM TYPE
#endif
#endif
	char path [MAX_PATHNAME+1];	/* Inter-process communication */
					/* file name */

	sprintf (path, "%s/%1d", Nqs_inter, pid);
#if	BSD42 | BSD43 | ULTRIX
	time (&utimbuf [0].tv_sec);	/* Access time = current time */
	utimbuf [0].tv_usec = 0;
	utimbuf [1].tv_sec = completion_code + LMTOFFSET;
	utimbuf [1].tv_usec = 0;
	if (utimes (path, utimbuf) == -1) {
#else
#if	UNICOS | SGI | SYS52 | SYS53 | UTS
	time (&utimbuf.acctime);	/* Access time = current time */
	utimbuf.modtime = completion_code;
	if (utime (path, &utimbuf) == -1) {
#else
BAD SYSTEM TYPE
#endif
#endif
		/*
		 *  We were not successful.
		 */
		if (errno != ENOENT) {
			/*
			 *  A very serious error has occurred; either
			 *
			 *	EPERM, ENOTDIR, EACCES, EROFS, EFAULT, or
			 *	ELOOP (Berkeley).
			 */
#if	BSD42 | BSD43 | ULTRIX
			printf ("F$Utimes() call error in nqs_complt().\n");
#else
#if	UNICOS | SGI | SYS52 | SYS53 | UTS
			printf ("F$Utime() call error in nqs_complt().\n");
#else
BAD SYSTEM TYPE
#endif
#endif
			nqs_abort();	/* Abort execution */
		}
		/*
		 *  The inter-process communication file did not exist.
		 *  Check to see if the client process still exists.
		 *  If the client does not exist, then we do not report
		 *  an error, and instead assume that the client has
		 *  exited prematurely, without removing their inter-
		 *  process communication file.
		 */
		if (kill (pid, 0) != -1) {
			/*
			 *  The client process still exists, but the
			 *  inter-process communication file isn't there,
			 *  or we were given the wrong process-id.
			 */
			printf ("E$Wrong pid or missing ");
			printf ("inter-process comm. file.\n");
			printf (incomplete);
			fflush (stdout);
		}
	}
	else if (
#ifdef	AIX
		 sleep (1) == -1 ||	/* Delay the signal ! */
#endif
		 kill (pid, SIGALRM) == -1) {
		/*
		 *  We successfully altered the access and modification
		 *  times of the inter-process communication file, but
		 *  were unable to signal the client process that we had
		 *  completed the transaction.
		 */
		if (errno == EINVAL) {
			printf ("E$Invalid pid given to nqs_complt().\n");
			printf (incomplete);
			fflush (stdout);
		}
		else if (errno == ESRCH) {
			/*
			 *  The client process no longer exists to
			 *  listen to our results.
			 */
			unlink (path);	/* Remove the inter-process */
					/* communication file for the */
					/* exited process. */
		}
		else if (errno == EPERM) {
			printf ("F$NQS not running as root in ");
			printf ("nqs_complt().\n");
			nqs_abort();
		}
	}
}
